# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/scilint.ipynb.

# %% auto 0
__all__ = ['run_nbqa_cmd', 'tidy', 'get_project_root', 'is_nbdev_project', 'get_default_spec', 'lint_nb', 'lint_nbs',
           'display_warning_report', 'lint', 'build', 'scilint_tidy', 'scilint_lint', 'scilint_build', 'scilint_ci']

# %% ../nbs/scilint.ipynb 2
import json
import operator
import os
import shutil
import sys
from configparser import InterpolationMissingOptionError
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Tuple

import numpy as np
import pandas as pd
import yaml
from execnb.nbio import read_nb
from fastcore.script import Param, call_parse, store_false
from fastcore.xtras import globtastic
from nbdev.clean import nbdev_clean
from nbdev.config import get_config
from nbdev.doclinks import nbdev_export, nbglob
from nbdev.quarto import nbdev_docs, nbdev_readme
from nbdev.test import nbdev_test
from nbqa.__main__ import _get_configs, _main
from nbqa.cmdline import CLIArgs
from nbqa.find_root import find_project_root

from .indicators import indicator_funcs

# %% ../nbs/scilint.ipynb 7
def run_nbqa_cmd(cmd):
    print(f"Running {cmd}")
    project_root: Path = find_project_root(tuple([str(Path(".").resolve())]))
    args = CLIArgs.parse_args([cmd, str(project_root)])
    configs = _get_configs(args, project_root)
    output_code = _main(args, configs)
    return output_code

# %% ../nbs/scilint.ipynb 9
def tidy():
    tidy_tools = ["black", "isort", "autoflake"]
    [run_nbqa_cmd(c) for c in tidy_tools]

# %% ../nbs/scilint.ipynb 11
def get_project_root(path: Path = Path(".").resolve()):
    return find_project_root(tuple([str()]))

# %% ../nbs/scilint.ipynb 12
def is_nbdev_project(project_path: Path = Path(".")):
    is_nbdev = True
    project_root = find_project_root(tuple([str(project_path.resolve())]))

    if not Path(project_root, "settings.ini").exists():
        is_nbdev = False
    try:
        get_config().lib_name
    except InterpolationMissingOptionError:
        is_nbdev = False

    return is_nbdev

# %% ../nbs/scilint.ipynb 15
def get_default_spec():
    return {
        "exclusions": None,
        "fail_over": 1,
        "out_dir": "/tmp/scilint/",
        "precision": 3,
        "print_syntax_errors": False,
        "evaluate": True,
        "warnings": {
            "lt": {
                "calls_per_func_median": 0,
                "calls_per_func_mean": 1,
                "in_func_pct": 20,
                "asserts_func_ratio": 1,
                "tests_func_coverage_pct": 20,
                "tests_per_func_mean": 0.5,
                "markdown_code_pct": 5,
            },
            "gt": {"total_code_len": 50000, "loc_per_md_section": 2000},
            "equals": {"has_syntax_error": True},
        },
    }

# %% ../nbs/scilint.ipynb 18
def lint_nb(
    spec_name: str,
    nb_path: Path,
    conf: Dict[str, Any],
    indicators: Dict[str, Callable],
    include_in_scoring: bool,
) -> Tuple[float]:
    nb = read_nb(nb_path)

    has_syntax_error = False
    indic_vals = list(np.repeat(np.nan, len(indicators)))
    try:
        for i, indic_name in enumerate(indicators):
            indic_vals[i] = round(indicators[indic_name](nb), conf["precision"])
    except SyntaxError as se:
        if conf["print_syntax_errors"]:
            print(f"Syntax error in notebook: {nb_path} reason: ", se)
        has_syntax_error = True
    indic_vals.append(has_syntax_error)
    indic_vals.append(include_in_scoring)
    indic_vals.insert(0, spec_name)

    return tuple(indic_vals)

# %% ../nbs/scilint.ipynb 21
def _get_excluded_paths(paths: Iterable[Path], exclude_pattern: str) -> Iterable[Path]:
    """Excluded paths should either be absolute paths or paths rooted at the project root directory"""
    excl_paths = []
    paths = [p.absolute() for p in paths]

    for ex_pattern in exclude_pattern.split(","):
        if Path(ex_pattern).is_absolute():
            ex_path = Path(ex_pattern)
        else:
            ex_path = Path(get_project_root(), ex_pattern)

        if ex_path.exists():
            excl_paths.extend([p for p in paths if ex_pattern in str(p)])
        elif not ex_path.exists():
            raise ValueError(f"Path component: {ex_path} does not exist")
        else:
            raise ValueError(
                f"Invalid exclusion pattern: {ex_path} pattern is comma separrated list of 'dir/' for directories and 'name.ipynb' for specific notebook"
            )
    return excl_paths

# %% ../nbs/scilint.ipynb 24
def _calculate_warnings(
    spec_name: str,
    scoring_report: pd.DataFrame,
    conf: Dict[str, Any],
    include_missing: bool = False,
) -> Tuple[Dict[str, Any], int]:
    warning_details = []

    for op_text in list(conf["warnings"].keys()):
        for indic in conf["warnings"][op_text]:
            metric_series = scoring_report[indic]
            or_exp = pd.isnull(metric_series) if include_missing else False
            op = (
                operator.lt
                if op_text == "lt"
                else operator.gt
                if op_text == "gt"
                else operator.eq
            )
            warning_data = metric_series[
                (op(metric_series, conf["warnings"][op_text][indic])) | (or_exp)
            ]
            warning_dict = warning_data.to_dict()
            for key, val in warning_dict.items():
                warning_dict[key] = (
                    indic,
                    val,
                    op_text,
                    conf["warnings"][op_text][indic],
                )
            warning_details.append(warning_dict)

    all_warns = _reshape_warnings(spec_name, scoring_report, warning_details)
    num_warnings = len(all_warns)
    return all_warns, num_warnings

# %% ../nbs/scilint.ipynb 27
def _reshape_warnings(
    spec_name: str, scoring_report: pd.DataFrame, warning_details: Iterable[Any]
) -> Dict[str, Iterable[Tuple]]:
    warnings_by_nb = {nb: [] for nb in scoring_report.index}
    for nb in scoring_report.index:
        for wd in warning_details:
            if nb in wd:
                warnings_by_nb[nb].append(tuple([spec_name, nb] + list(wd[nb])))
    warnings_by_nb = {key: val for key, val in warnings_by_nb.items() if len(val) > 0}
    flattened_warns = [item for sublist in warnings_by_nb.values() for item in sublist]
    return pd.DataFrame.from_records(
        data=flattened_warns,
        columns=[
            "spec_name",
            "notebook",
            "indicator",
            "value",
            "operator",
            "threshold",
        ],
    )

# %% ../nbs/scilint.ipynb 30
def lint_nbs(
    spec_name: str,
    conf: Dict[str, Any],
    indicators: Dict[str, Callable],
    nb_paths: Iterable[Path] = None,
    nb_glob: Path = None,
):
    if nb_paths is None:
        nb_paths = [Path(p).absolute() for p in nbglob(nb_glob)]
    else:
        nb_paths = [Path(p).absolute() for p in nb_paths]

    if len(nb_paths) == 0:
        return None, None, None

    excluded_paths = None
    exclusions = conf["exclusions"]
    if exclusions is not None:
        excluded_paths = _get_excluded_paths(nb_paths, exclude_pattern=exclusions)

    results = []
    nb_names = []
    for nb_path in nb_paths:
        include_in_scoring = True
        if exclusions is not None:
            include_in_scoring = False if nb_path in excluded_paths else True

        nb_names.append(nb_path.stem)
        lint_result = lint_nb(spec_name, nb_path, conf, indicators, include_in_scoring)
        results.append(lint_result)

    lint_report = pd.DataFrame.from_records(
        data=results,
        index=nb_names,
        columns=["spec_name"]
        + list(indicators.keys())
        + ["has_syntax_error", "include_in_scoring"],
    ).sort_values(["in_func_pct", "markdown_code_pct"], ascending=False)

    scoring_report = lint_report[lint_report.include_in_scoring].copy()
    all_warns, num_warnings = _calculate_warnings(spec_name, scoring_report, conf)
    return lint_report, all_warns, num_warnings

# %% ../nbs/scilint.ipynb 32
def _map_paths_specs(nb_glob: Path = None, specs_glob: Path = Path(".").resolve()):
    nbs = nbglob(nb_glob)
    spec_files = [
        Path(p)
        for p in globtastic(
            specs_glob,
            file_glob="scilint-*.yaml",
            skip_folder_re="ipynb_checkpoints|_proc",
        )
    ]
    default_spec_files = [p for p in spec_files if p.name == "scilint-default.yaml"]
    default_spec_file = default_spec_files[0] if len(default_spec_files) > 0 else None
    spec_dirs = [p.parent for p in spec_files]

    spec_nbs = {k: [] for k in spec_files}
    for nb in [Path(p) for p in nbs]:
        found_spec = False
        for name, spec_dir in zip(spec_files, spec_dirs):
            if nb.parent == spec_dir:
                spec_nbs[name].append(nb)
                found_spec = True
        if not found_spec:
            if default_spec_file is not None:
                spec_nbs[default_spec_file].append(nb)
            else:
                # Special case: not actually a valid file path - triggers loading a fallback
                fallback_path = Path("scilint-default")
                if fallback_path not in spec_nbs:
                    spec_nbs[fallback_path] = []
                spec_nbs[fallback_path].append(nb)

    return spec_nbs

# %% ../nbs/scilint.ipynb 47
def display_warning_report(all_warns: pd.DataFrame):
    print(
        "\n******************************************Begin Scilint Warning Report*****************************************"
    )
    print(all_warns.to_markdown(tablefmt="grid", index=False))
    print(
        "\n******************************************End Scilint Warning Report*******************************************\n"
    )

# %% ../nbs/scilint.ipynb 49
def _persist_results(
    lint_report: pd.DataFrame, all_warns: pd.DataFrame, conf: Dict[str, Any]
):
    out_dir = Path(conf["out_dir"])
    conf_to_persist = {k: v for k, v in conf.items() if k != "indicators"}
    if not out_dir.exists():
        Path(out_dir).mkdir()
    with open(Path(out_dir, "scilint_config.json"), "w") as outfile:
        json.dump(conf_to_persist, outfile)
    all_warns.to_csv(Path(out_dir, "scilint_warnings.csv"), index=False)
    lint_report.to_csv(Path(out_dir, "scilint_report.csv"))

# %% ../nbs/scilint.ipynb 52
def _load_conf(
    conf_path: str = None,
    exclusions: str = None,
    fail_over: int = None,
    out_dir: int = None,
    precision: int = None,
    print_syntax_errors: bool = None,
):
    if conf_path is None:
        project_root = find_project_root(tuple([str(Path(".").resolve())]))
        conf_path = Path(project_root, "nbs", "scilint-default.yaml")
        print(f"Loading default lint config: {conf_path}")
    else:
        conf_path = Path(conf_path)

    conf = yaml.safe_load(conf_path.read_text())
    override_names = (
        "exclusions",
        "fail_over",
        "out_dir",
        "precision",
        "print_syntax_errors",
    )
    overrides = (exclusions, fail_over, out_dir, precision, print_syntax_errors)
    for override in zip(override_names, overrides):
        if override[1] is not None:
            conf[override[0]] = override[1]
    return conf

# %% ../nbs/scilint.ipynb 56
def lint(
    display_report: bool = True,
    nb_glob: Path = None,
    specs_glob: Path = Path(".").resolve(),
    exclusions: str = None,
    fail_over: int = None,
    out_dir: int = None,
    precision: int = None,
    print_syntax_errors: bool = None,
):
    spec_nbs = _map_paths_specs(nb_glob, specs_glob)
    lint_reports = []
    all_warns = []
    warns_count = []
    for spec, nbs in spec_nbs.items():
        if spec == "scilint-default":
            conf = get_default_spec()
        else:
            conf = _load_conf(
                spec, exclusions, fail_over, out_dir, precision, print_syntax_errors
            )
        if conf["evaluate"] == False:
            print(f"Linting skipped for: {spec.name} as evaluate is set to false")
            continue
        lint_report, report_warns, num_warnings = lint_nbs(
            spec.name, conf, indicator_funcs, nb_paths=nbs
        )
        lint_reports.append(lint_report)
        all_warns.append(report_warns)
        warns_count.append(num_warnings)

        fail_over_conf = conf["fail_over"]
        if conf["fail_over"] == -1:
            print(f"Linting warnings ignored for: {spec.name} as fail_over set to -1")
        elif num_warnings == 0:
            print(f"Linting success for: {spec.name}, no issues found")
        elif num_warnings <= conf["fail_over"]:
            print(
                f"Linting success for: {spec.name}, warnings ({num_warnings}) <= than threshold ({fail_over_conf}) "
            )
        else:
            print(
                f"Linting failed for: {spec.name}, total warnings ({num_warnings}) exceeded threshold ({fail_over_conf})"
            )
            sys.exit(num_warnings)

    lint_report = pd.concat(lint_reports) if len(lint_reports) > 0 else lint_report
    all_warns = pd.concat(all_warns) if len(all_warns) > 0 else report_warns
    num_warnings = sum(warns_count)

    if num_warnings > 0:
        print(
            f"{num_warnings} warnings founds, within tolerated thresholds for all specs"
        )
        if display_report:
            display_warning_report(all_warns)
    elif num_warnings == 0:
        print("No issues found during linting")

    _persist_results(lint_report, all_warns, conf)
    print("Linting completed")

# %% ../nbs/scilint.ipynb 59
def build(
    display_report: bool = True,
    nb_glob: Path = None,
    specs_glob: Path = Path(".").resolve(),
    exclusions: str = None,
    fail_over: int = None,
    out_dir: int = None,
    precision: int = None,
    print_syntax_errors: bool = None,
):
    print("Tidying notebooks..")
    tidy()
    if is_nbdev_project():
        nbdev_export.__wrapped__()
        print("Converted notebooks to modules")
        print("Testing notebooks..")
        nbdev_test.__wrapped__()
    print("Running notebook linter..")
    lint(
        display_report,
        nb_glob,
        specs_glob,
        exclusions,
        fail_over,
        out_dir,
        precision,
        print_syntax_errors,
    )
    if is_nbdev_project():
        nbdev_clean.__wrapped__()
        print("Cleaned notebooks")

# %% ../nbs/scilint.ipynb 62
@call_parse
def scilint_tidy():
    tidy()

# %% ../nbs/scilint.ipynb 64
@call_parse
def scilint_lint(
    display_report: Param("Print the lint report", store_false) = False,
    nb_glob: Path = None,
    specs_glob: Path = Path(".").resolve(),
    exclusions: str = None,
    fail_over: int = None,
    out_dir: int = None,
    precision: int = None,
    print_syntax_errors: bool = None,
):
    lint(
        display_report,
        nb_glob,
        specs_glob,
        exclusions,
        fail_over,
        out_dir,
        precision,
        print_syntax_errors,
    )

# %% ../nbs/scilint.ipynb 67
@call_parse
def scilint_build(
    display_report: Param("Print the lint report", store_false) = False,
    nb_glob: Path = None,
    specs_glob: Path = Path(".").resolve(),
    exclusions: str = None,
    fail_over: int = None,
    out_dir: int = None,
    precision: int = None,
    print_syntax_errors: bool = None,
):
    build(
        display_report,
        nb_glob,
        specs_glob,
        exclusions,
        fail_over,
        out_dir,
        precision,
        print_syntax_errors,
    )

# %% ../nbs/scilint.ipynb 69
@call_parse
def scilint_ci(
    display_report: Param("Print the lint report", store_false) = False,
    nb_glob: Path = None,
    specs_glob: Path = Path(".").resolve(),
    exclusions: str = None,
    fail_over: int = None,
    out_dir: int = None,
    precision: int = None,
    print_syntax_errors: bool = None,
):
    if not is_nbdev_project():
        print("scilint_ci feature is only available for nbdev projects")
        return

    build(
        display_report,
        nb_glob,
        specs_glob,
        exclusions,
        fail_over,
        out_dir,
        precision,
        print_syntax_errors,
    )

    if not shutil.which("quarto"):
        print(
            "Quarto is not installed. A working quarto install is required for the CI build"
        )
        sys.exit(-1)
    nbdev_readme.__wrapped__()
    nbdev_docs.__wrapped__()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb.

# %% auto 0
__all__ = ['traffic_percent', 'workers', 'model_level', 'min_date', 'something', 'get_traffic_text', 'get_experiment_segment',
           'get_utterances', 'preprocess', 'Topics', 'fit', 'evaluate', 'serve_num_topics']

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 2
def something():
    message = "The first step"
    print(f"{message}")
    results = {"message": message}
    return results

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 6
import numpy as np
import pandas as pd

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 8
traffic_percent = 1
workers = 8
model_level = "dispatcher"
min_date = "2021-01-01"

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 9
def get_traffic_text(percent):
    return str(percent) if int(percent) >= 10 else "0" + str(percent)

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 15
def get_experiment_segment(traffic_percent):
    return tuple(get_traffic_text(tp) for tp in range(traffic_percent))

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 17
def get_utterances(model_level=None, min_date=None, traffic_percent=100):
    """
    You will probably call data preparation code here. To simplify dependencies we are just creating synthetic data instead.
    """
    get_experiment_segment(traffic_percent)
    dummy_data = pd.Series(
        np.random.choice(
            [
                "Hello",
                "Goodbye",
                "Hi",
                "Can you help?",
                "I have an issue, can you help me?",
            ],
            100,
        ),
        name="utterance",
    )
    return dummy_data

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 18
def preprocess(message, model_level=None, min_date=None, traffic_percent=100):
    print(f"I captialised the message: {message.upper()}")
    data = get_utterances(model_level, min_date, traffic_percent)
    documents = data.tolist()
    results = {"documents": documents}
    return results

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 21
class Topics:
    def __init__(self, documents, workers):
        pass

    def get_num_topics(self):
        return 6

    def get_topic_sizes(self):
        return [1, 2, 3, 4, 5, 6], [1, 2, 3, 4, 5, 6]

    def get_topics(self, num_topics):
        return (
            ["cat", "sat", "mat", "mouse", "house", "grouse"],
            np.asarray([1, 1, 1, 1, 1, 1]),
            [1, 2, 3, 4, 5, 6],
        )

    def plot_wordcloud(self):
        print("you may want to remove plotting code from testing to speed things up")

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 23
def fit(documents, workers=workers):
    model = Topics(documents, workers=workers)
    results = {"model": model}
    return results

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 35
def evaluate(model):
    topic_words, word_scores, topic_nums = model.get_topics(model.get_num_topics())

    topic_contains_non_empty_words = all([len(tw) > 0 for tw in topic_words])
    word_scores_in_range = word_scores.min() >= 0.0 and word_scores.max() <= 1.0
    as_many_items_as_topics = (
        model.get_num_topics() == len(topic_words) == word_scores.shape[0]
    )
    word_summaries = (
        topic_contains_non_empty_words
        and word_scores_in_range
        and as_many_items_as_topics
    )
    # You can add artifacts in a step that will be saved to block storage. Add the paths to the file on the local filesystem
    # and the artifact will be uploaded to remote storage.
    sample_df = pd.DataFrame(
        {"a": model.get_topic_sizes()[0], "b": model.get_topic_sizes()[1]}
    )
    sample_df.to_csv("/tmp/dataframe_artifact.csv", index=False)
    artifacts = ["/tmp/dataframe_artifact.csv"]
    # You can add step metrics too this time just add a list of 3-tuples where tuple order = (name, value, step)
    metrics = [("mae", 100, 0), ("mae", 67, 1), ("mae", 32, 2)]
    results = {
        "word_summaries": word_summaries,
        "artifacts": artifacts,
        "metrics": metrics,
    }
    return results

# %% ../../nbs/example_nbs/no_spec_provided/nbdev.ipynb 38
def serve_num_topics(model):
    return model.get_num_topics()



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/scilint/',
                'doc_host': 'https://newday-data.github.io',
                'git_url': 'https://github.com/newday-data/scilint/tree/{branch}/',
                'lib_path': 'scilint'},
  'syms': { 'scilint.experimental.test_nbdev': { 'scilint.experimental.test_nbdev.Topics': ( 'example_nbs/experimental/nbdev.html#topics',
                                                                                             'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.Topics.__init__': ( 'example_nbs/experimental/nbdev.html#topics.__init__',
                                                                                                      'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.Topics.get_num_topics': ( 'example_nbs/experimental/nbdev.html#topics.get_num_topics',
                                                                                                            'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.Topics.get_topic_sizes': ( 'example_nbs/experimental/nbdev.html#topics.get_topic_sizes',
                                                                                                             'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.Topics.get_topics': ( 'example_nbs/experimental/nbdev.html#topics.get_topics',
                                                                                                        'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.Topics.plot_wordcloud': ( 'example_nbs/experimental/nbdev.html#topics.plot_wordcloud',
                                                                                                            'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.evaluate': ( 'example_nbs/experimental/nbdev.html#evaluate',
                                                                                               'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.fit': ( 'example_nbs/experimental/nbdev.html#fit',
                                                                                          'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.get_experiment_segment': ( 'example_nbs/experimental/nbdev.html#get_experiment_segment',
                                                                                                             'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.get_traffic_text': ( 'example_nbs/experimental/nbdev.html#get_traffic_text',
                                                                                                       'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.get_utterances': ( 'example_nbs/experimental/nbdev.html#get_utterances',
                                                                                                     'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.preprocess': ( 'example_nbs/experimental/nbdev.html#preprocess',
                                                                                                 'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.serve_num_topics': ( 'example_nbs/experimental/nbdev.html#serve_num_topics',
                                                                                                       'scilint/experimental/test_nbdev.py'),
                                                 'scilint.experimental.test_nbdev.something': ( 'example_nbs/experimental/nbdev.html#something',
                                                                                                'scilint/experimental/test_nbdev.py')},
            'scilint.exploratory.test_nbdev': { 'scilint.exploratory.test_nbdev.Topics': ( 'example_nbs/exploratory/nbdev.html#topics',
                                                                                           'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.Topics.__init__': ( 'example_nbs/exploratory/nbdev.html#topics.__init__',
                                                                                                    'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.Topics.get_num_topics': ( 'example_nbs/exploratory/nbdev.html#topics.get_num_topics',
                                                                                                          'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.Topics.get_topic_sizes': ( 'example_nbs/exploratory/nbdev.html#topics.get_topic_sizes',
                                                                                                           'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.Topics.get_topics': ( 'example_nbs/exploratory/nbdev.html#topics.get_topics',
                                                                                                      'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.Topics.plot_wordcloud': ( 'example_nbs/exploratory/nbdev.html#topics.plot_wordcloud',
                                                                                                          'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.evaluate': ( 'example_nbs/exploratory/nbdev.html#evaluate',
                                                                                             'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.fit': ( 'example_nbs/exploratory/nbdev.html#fit',
                                                                                        'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.get_experiment_segment': ( 'example_nbs/exploratory/nbdev.html#get_experiment_segment',
                                                                                                           'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.get_traffic_text': ( 'example_nbs/exploratory/nbdev.html#get_traffic_text',
                                                                                                     'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.get_utterances': ( 'example_nbs/exploratory/nbdev.html#get_utterances',
                                                                                                   'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.preprocess': ( 'example_nbs/exploratory/nbdev.html#preprocess',
                                                                                               'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.serve_num_topics': ( 'example_nbs/exploratory/nbdev.html#serve_num_topics',
                                                                                                     'scilint/exploratory/test_nbdev.py'),
                                                'scilint.exploratory.test_nbdev.something': ( 'example_nbs/exploratory/nbdev.html#something',
                                                                                              'scilint/exploratory/test_nbdev.py')},
            'scilint.indicators': { 'scilint.indicators._check_for_function_asserts': ( 'indicators.html#_check_for_function_asserts',
                                                                                        'scilint/indicators.py'),
                                    'scilint.indicators._count_func_ret_asserts': ( 'indicators.html#_count_func_ret_asserts',
                                                                                    'scilint/indicators.py'),
                                    'scilint.indicators._count_inline_asserts': ( 'indicators.html#_count_inline_asserts',
                                                                                  'scilint/indicators.py'),
                                    'scilint.indicators._incr_assert_count': ( 'indicators.html#_incr_assert_count',
                                                                               'scilint/indicators.py'),
                                    'scilint.indicators._tests_per_function_code': ( 'indicators.html#_tests_per_function_code',
                                                                                     'scilint/indicators.py'),
                                    'scilint.indicators._update_ret_vals': ('indicators.html#_update_ret_vals', 'scilint/indicators.py'),
                                    'scilint.indicators.calc_ifp': ('indicators.html#calc_ifp', 'scilint/indicators.py'),
                                    'scilint.indicators.calls_per_func': ('indicators.html#calls_per_func', 'scilint/indicators.py'),
                                    'scilint.indicators.calls_per_func_mean': ( 'indicators.html#calls_per_func_mean',
                                                                                'scilint/indicators.py'),
                                    'scilint.indicators.calls_per_func_median': ( 'indicators.html#calls_per_func_median',
                                                                                  'scilint/indicators.py'),
                                    'scilint.indicators.count_func_calls': ('indicators.html#count_func_calls', 'scilint/indicators.py'),
                                    'scilint.indicators.get_cell_code': ('indicators.html#get_cell_code', 'scilint/indicators.py'),
                                    'scilint.indicators.get_func_defs': ('indicators.html#get_func_defs', 'scilint/indicators.py'),
                                    'scilint.indicators.get_project_root': ('indicators.html#get_project_root', 'scilint/indicators.py'),
                                    'scilint.indicators.iaf': ('indicators.html#iaf', 'scilint/indicators.py'),
                                    'scilint.indicators.in_func_pct': ('indicators.html#in_func_pct', 'scilint/indicators.py'),
                                    'scilint.indicators.loc_per_md_section': ( 'indicators.html#loc_per_md_section',
                                                                               'scilint/indicators.py'),
                                    'scilint.indicators.markdown_code_pct': ('indicators.html#markdown_code_pct', 'scilint/indicators.py'),
                                    'scilint.indicators.replace_ipython_magics': ( 'indicators.html#replace_ipython_magics',
                                                                                   'scilint/indicators.py'),
                                    'scilint.indicators.safe_div': ('indicators.html#safe_div', 'scilint/indicators.py'),
                                    'scilint.indicators.tests_func_coverage_pct': ( 'indicators.html#tests_func_coverage_pct',
                                                                                    'scilint/indicators.py'),
                                    'scilint.indicators.tests_per_func_mean': ( 'indicators.html#tests_per_func_mean',
                                                                                'scilint/indicators.py'),
                                    'scilint.indicators.tests_per_function': ( 'indicators.html#tests_per_function',
                                                                               'scilint/indicators.py'),
                                    'scilint.indicators.total_code_len': ('indicators.html#total_code_len', 'scilint/indicators.py')},
            'scilint.scilint': { 'scilint.scilint._calculate_warnings': ('scilint.html#_calculate_warnings', 'scilint/scilint.py'),
                                 'scilint.scilint._get_excluded_paths': ('scilint.html#_get_excluded_paths', 'scilint/scilint.py'),
                                 'scilint.scilint._load_conf': ('scilint.html#_load_conf', 'scilint/scilint.py'),
                                 'scilint.scilint._map_paths_specs': ('scilint.html#_map_paths_specs', 'scilint/scilint.py'),
                                 'scilint.scilint._persist_results': ('scilint.html#_persist_results', 'scilint/scilint.py'),
                                 'scilint.scilint._reshape_warnings': ('scilint.html#_reshape_warnings', 'scilint/scilint.py'),
                                 'scilint.scilint.build': ('scilint.html#build', 'scilint/scilint.py'),
                                 'scilint.scilint.display_warning_report': ('scilint.html#display_warning_report', 'scilint/scilint.py'),
                                 'scilint.scilint.get_default_spec': ('scilint.html#get_default_spec', 'scilint/scilint.py'),
                                 'scilint.scilint.get_project_root': ('scilint.html#get_project_root', 'scilint/scilint.py'),
                                 'scilint.scilint.is_nbdev_project': ('scilint.html#is_nbdev_project', 'scilint/scilint.py'),
                                 'scilint.scilint.lint': ('scilint.html#lint', 'scilint/scilint.py'),
                                 'scilint.scilint.lint_nb': ('scilint.html#lint_nb', 'scilint/scilint.py'),
                                 'scilint.scilint.lint_nbs': ('scilint.html#lint_nbs', 'scilint/scilint.py'),
                                 'scilint.scilint.run_nbqa_cmd': ('scilint.html#run_nbqa_cmd', 'scilint/scilint.py'),
                                 'scilint.scilint.scilint_build': ('scilint.html#scilint_build', 'scilint/scilint.py'),
                                 'scilint.scilint.scilint_ci': ('scilint.html#scilint_ci', 'scilint/scilint.py'),
                                 'scilint.scilint.scilint_lint': ('scilint.html#scilint_lint', 'scilint/scilint.py'),
                                 'scilint.scilint.scilint_tidy': ('scilint.html#scilint_tidy', 'scilint/scilint.py'),
                                 'scilint.scilint.tidy': ('scilint.html#tidy', 'scilint/scilint.py')},
            'scilint.test.test_nbdev': { 'scilint.test.test_nbdev.Topics': ( 'example_nbs/no_spec_provided/nbdev.html#topics',
                                                                             'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.Topics.__init__': ( 'example_nbs/no_spec_provided/nbdev.html#topics.__init__',
                                                                                      'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.Topics.get_num_topics': ( 'example_nbs/no_spec_provided/nbdev.html#topics.get_num_topics',
                                                                                            'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.Topics.get_topic_sizes': ( 'example_nbs/no_spec_provided/nbdev.html#topics.get_topic_sizes',
                                                                                             'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.Topics.get_topics': ( 'example_nbs/no_spec_provided/nbdev.html#topics.get_topics',
                                                                                        'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.Topics.plot_wordcloud': ( 'example_nbs/no_spec_provided/nbdev.html#topics.plot_wordcloud',
                                                                                            'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.evaluate': ( 'example_nbs/no_spec_provided/nbdev.html#evaluate',
                                                                               'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.fit': ( 'example_nbs/no_spec_provided/nbdev.html#fit',
                                                                          'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.get_experiment_segment': ( 'example_nbs/no_spec_provided/nbdev.html#get_experiment_segment',
                                                                                             'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.get_traffic_text': ( 'example_nbs/no_spec_provided/nbdev.html#get_traffic_text',
                                                                                       'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.get_utterances': ( 'example_nbs/no_spec_provided/nbdev.html#get_utterances',
                                                                                     'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.preprocess': ( 'example_nbs/no_spec_provided/nbdev.html#preprocess',
                                                                                 'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.serve_num_topics': ( 'example_nbs/no_spec_provided/nbdev.html#serve_num_topics',
                                                                                       'scilint/test/test_nbdev.py'),
                                         'scilint.test.test_nbdev.something': ( 'example_nbs/no_spec_provided/nbdev.html#something',
                                                                                'scilint/test/test_nbdev.py')},
            'scilint.test.test_nbdev_high_quality': { 'scilint.test.test_nbdev_high_quality.Topics': ( 'example_nbs/nbdev_high_quality.html#topics',
                                                                                                       'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.Topics.__init__': ( 'example_nbs/nbdev_high_quality.html#topics.__init__',
                                                                                                                'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.Topics.get_num_topics': ( 'example_nbs/nbdev_high_quality.html#topics.get_num_topics',
                                                                                                                      'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.Topics.get_topic_sizes': ( 'example_nbs/nbdev_high_quality.html#topics.get_topic_sizes',
                                                                                                                       'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.Topics.get_topics': ( 'example_nbs/nbdev_high_quality.html#topics.get_topics',
                                                                                                                  'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.Topics.plot_wordcloud': ( 'example_nbs/nbdev_high_quality.html#topics.plot_wordcloud',
                                                                                                                      'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.echo': ( 'example_nbs/nbdev_high_quality.html#echo',
                                                                                                     'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.evaluate': ( 'example_nbs/nbdev_high_quality.html#evaluate',
                                                                                                         'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.fit': ( 'example_nbs/nbdev_high_quality.html#fit',
                                                                                                    'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.get_experiment_segment': ( 'example_nbs/nbdev_high_quality.html#get_experiment_segment',
                                                                                                                       'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.get_traffic_text': ( 'example_nbs/nbdev_high_quality.html#get_traffic_text',
                                                                                                                 'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.get_utterances': ( 'example_nbs/nbdev_high_quality.html#get_utterances',
                                                                                                               'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.preprocess': ( 'example_nbs/nbdev_high_quality.html#preprocess',
                                                                                                           'scilint/test/test_nbdev_high_quality.py'),
                                                      'scilint.test.test_nbdev_high_quality.serve_num_topics': ( 'example_nbs/nbdev_high_quality.html#serve_num_topics',
                                                                                                                 'scilint/test/test_nbdev_high_quality.py')},
            'scilint.validated.test_nbdev_high_quality': { 'scilint.validated.test_nbdev_high_quality.Topics': ( 'example_nbs/validated/nbdev_high_quality.html#topics',
                                                                                                                 'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.Topics.__init__': ( 'example_nbs/validated/nbdev_high_quality.html#topics.__init__',
                                                                                                                          'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.Topics.get_num_topics': ( 'example_nbs/validated/nbdev_high_quality.html#topics.get_num_topics',
                                                                                                                                'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.Topics.get_topic_sizes': ( 'example_nbs/validated/nbdev_high_quality.html#topics.get_topic_sizes',
                                                                                                                                 'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.Topics.get_topics': ( 'example_nbs/validated/nbdev_high_quality.html#topics.get_topics',
                                                                                                                            'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.Topics.plot_wordcloud': ( 'example_nbs/validated/nbdev_high_quality.html#topics.plot_wordcloud',
                                                                                                                                'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.echo': ( 'example_nbs/validated/nbdev_high_quality.html#echo',
                                                                                                               'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.evaluate': ( 'example_nbs/validated/nbdev_high_quality.html#evaluate',
                                                                                                                   'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.fit': ( 'example_nbs/validated/nbdev_high_quality.html#fit',
                                                                                                              'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.get_experiment_segment': ( 'example_nbs/validated/nbdev_high_quality.html#get_experiment_segment',
                                                                                                                                 'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.get_traffic_text': ( 'example_nbs/validated/nbdev_high_quality.html#get_traffic_text',
                                                                                                                           'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.get_utterances': ( 'example_nbs/validated/nbdev_high_quality.html#get_utterances',
                                                                                                                         'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.preprocess': ( 'example_nbs/validated/nbdev_high_quality.html#preprocess',
                                                                                                                     'scilint/validated/test_nbdev_high_quality.py'),
                                                           'scilint.validated.test_nbdev_high_quality.serve_num_topics': ( 'example_nbs/validated/nbdev_high_quality.html#serve_num_topics',
                                                                                                                           'scilint/validated/test_nbdev_high_quality.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/indicators.ipynb.

# %% auto 0
__all__ = ['indicator_funcs', 'get_project_root', 'count_func_calls', 'replace_ipython_magics', 'safe_div', 'get_cell_code',
           'get_func_defs', 'calls_per_func', 'calls_per_func_mean', 'calls_per_func_median', 'iaf',
           'tests_per_function', 'tests_per_func_mean', 'tests_func_coverage_pct', 'calc_ifp', 'in_func_pct',
           'markdown_code_pct', 'total_code_len', 'loc_per_md_section']

# %% ../nbs/indicators.ipynb 2
import ast
import re
import warnings
from collections import Counter
from pathlib import Path

import nbformat
import numpy as np
import pandas as pd
from execnb.nbio import read_nb

# %% ../nbs/indicators.ipynb 7
def get_project_root(path: Path = Path(".").resolve()):
    return find_project_root(tuple([str()]))

# %% ../nbs/indicators.ipynb 9
def count_func_calls(code, func_defs):
    func_calls = Counter({k: 0 for k in func_defs})
    for stmt in ast.walk(ast.parse(code)):
        if isinstance(stmt, ast.Call):
            if type(stmt.func) == ast.Subscript:
                func_name = stmt.func.value.id
            else:
                func_name = (
                    stmt.func.id if "id" in stmt.func.__dict__ else stmt.func.attr
                )
            if func_name in func_defs:
                if func_name in func_calls:
                    func_calls[func_name] += 1
    return func_calls

# %% ../nbs/indicators.ipynb 13
def replace_ipython_magics(code):
    # Replace Ipython magic and shell command symbol with comment
    code = code.replace("%", "#")
    code = re.sub(r"^!", "#", code)
    return re.sub(r"\n\W?!", "\n#", code)

# %% ../nbs/indicators.ipynb 15
def safe_div(numer, denom):
    return 0 if denom == 0 else numer / denom

# %% ../nbs/indicators.ipynb 18
def get_cell_code(nb):
    pnb = nbformat.from_dict(nb)
    nb_cell_code = "\n".join(
        [
            replace_ipython_magics(c["source"])
            for c in pnb.cells
            if c["cell_type"] == "code"
        ]
    )
    return nb_cell_code

# %% ../nbs/indicators.ipynb 20
def get_func_defs(code, ignore_private_prefix=True):
    func_names = []
    for stmt in ast.walk(ast.parse(code)):
        if isinstance(stmt, ast.FunctionDef):
            inner_cond = (
                False if ignore_private_prefix and stmt.name.startswith("_") else True
            )
            if inner_cond:
                func_names.append(stmt.name)
    return func_names

# %% ../nbs/indicators.ipynb 24
def calls_per_func(nb):
    nb_cell_code = get_cell_code(nb)
    func_defs = get_func_defs(nb_cell_code)
    func_calls = count_func_calls(nb_cell_code, func_defs)
    return func_calls

# %% ../nbs/indicators.ipynb 26
def calls_per_func_mean(nb):
    return pd.Series(calls_per_func(nb)).mean()

# %% ../nbs/indicators.ipynb 28
def calls_per_func_median(nb):
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", message="Mean of empty slice")
        return pd.Series(calls_per_func(nb)).median()

# %% ../nbs/indicators.ipynb 37
def _count_inline_asserts(code, func_defs):
    inline_func_asserts = Counter({k: 0 for k in func_defs})

    for stmt in ast.walk(ast.parse(code)):
        if isinstance(stmt, ast.Assert):
            for assert_st in ast.walk(stmt):
                if isinstance(assert_st, ast.Call):
                    func_name = (
                        assert_st.func.id
                        if "id" in assert_st.func.__dict__
                        else assert_st.func.attr
                    )
                    if func_name in inline_func_asserts:
                        inline_func_asserts[func_name] += 1
    return inline_func_asserts

# %% ../nbs/indicators.ipynb 38
def iaf(nb):
    nb_cell_code = get_cell_code(nb)
    if nb_cell_code == "":
        return np.nan
    func_defs = get_func_defs(nb_cell_code)
    return _count_inline_asserts(nb_cell_code, func_defs)

# %% ../nbs/indicators.ipynb 44
def _count_func_ret_asserts(nb_cell_code):
    ret_vals = {}
    func_defs = get_func_defs(nb_cell_code)
    func_ret_asserts = Counter({k: 0 for k in func_defs})
    assert_func_counts = {}
    for stmt in ast.walk(ast.parse(nb_cell_code)):
        if isinstance(stmt, ast.Assign) and isinstance(stmt.value, ast.Call):
            _update_ret_vals(stmt, ret_vals)

        if isinstance(stmt, ast.Assert):
            assert_func_counts[id(stmt)] = []
            _check_for_function_asserts(
                stmt, ret_vals, func_ret_asserts, assert_func_counts
            )

    return func_ret_asserts

# %% ../nbs/indicators.ipynb 45
def _check_for_function_asserts(
    stmt: ast.AST, ret_vals, func_ret_asserts, assert_func_counts
):
    if "left" in stmt.test.__dict__:
        if "id" in stmt.test.left.__dict__:
            _incr_assert_count(
                id(stmt),
                ret_vals,
                func_ret_asserts,
                assert_func_counts,
                stmt.test.left.id,
            )
        for comp in stmt.test.comparators:
            if "id" in comp.__dict__:
                _incr_assert_count(
                    id(stmt), ret_vals, func_ret_asserts, assert_func_counts, comp.id
                )
    elif isinstance(stmt.test, ast.Name):
        if "id" in stmt.test.__dict__:
            _incr_assert_count(
                id(stmt), ret_vals, func_ret_asserts, assert_func_counts, stmt.test.id
            )
    elif isinstance(stmt.test, ast.BoolOp):
        for val in stmt.test.values:
            if "left" in val.__dict__:
                if "id" in val.left.__dict__:
                    _incr_assert_count(
                        id(stmt),
                        ret_vals,
                        func_ret_asserts,
                        assert_func_counts,
                        val.left.id,
                    )
                for comp in val.comparators:
                    if "id" in comp.__dict__:
                        _incr_assert_count(
                            id(stmt),
                            ret_vals,
                            func_ret_asserts,
                            assert_func_counts,
                            comp.id,
                        )

# %% ../nbs/indicators.ipynb 46
def _incr_assert_count(
    assert_id, ret_vals, func_ret_asserts, assert_func_counts, return_var
):
    if (
        return_var in ret_vals
        and ret_vals[return_var] not in assert_func_counts[assert_id]
    ):
        assert_func_counts[assert_id].append(ret_vals[return_var])
        if return_var in ret_vals:
            func_ret_asserts[ret_vals[return_var]] += 1

# %% ../nbs/indicators.ipynb 47
def _update_ret_vals(stmt, ret_vals):
    if type(stmt.value.func) == ast.Subscript:
        func_name = stmt.func.value.id
    else:
        if isinstance(stmt.value.func, ast.Attribute):
            if (
                "func" in stmt.value.func.value.__dict__
                and "id" in stmt.value.func.value.func.__dict__
            ):
                func_name = stmt.value.func.value.func.id
            else:
                if "attr" in stmt.value.func.__dict__:
                    func_name = stmt.value.func.attr
                elif "value" in stmt.value.func.__dict__:
                    if "id" in stmt.value.func.value.__dict__:
                        func_name = stmt.value.func.value.id
                    elif "attr" in stmt.value.func.value.__dict__:
                        func_name = stmt.value.func.value.attr
                else:
                    raise ValueError(f"{stmt.value.func.__dict__}")
        else:
            func_name = (
                stmt.value.func.id
                if "id" in stmt.value.func.__dict__
                else stmt.func.attr
            )
    if isinstance(stmt.targets[0], ast.Name):
        ret_vals[stmt.targets[0].id] = func_name
    elif isinstance(stmt.targets[0], ast.Tuple):
        for elts in stmt.targets[0].elts:
            ret_vals[elts.id] = func_name

# %% ../nbs/indicators.ipynb 49
def tests_per_function(nb):
    nb_cell_code = "\n".join(
        [
            replace_ipython_magics(c["source"])
            for c in nb.cells
            if c["cell_type"] == "code"
        ]
    )
    return _tests_per_function_code(nb_cell_code)


def _tests_per_function_code(nb_cell_code):
    func_ret_asserts = _count_func_ret_asserts(nb_cell_code)
    inline_asserts = _count_inline_asserts(nb_cell_code, get_func_defs(nb_cell_code))

    func_ret_asserts.update(inline_asserts)
    return pd.Series(func_ret_asserts)

# %% ../nbs/indicators.ipynb 52
def tests_per_func_mean(nb):
    return tests_per_function(nb).mean()

# %% ../nbs/indicators.ipynb 54
def tests_func_coverage_pct(nb):
    return tests_per_function(nb).clip(upper=1).mean() * 100

# %% ../nbs/indicators.ipynb 59
def calc_ifp(nb_cell_code):
    stmts_in_func = 0
    stmts_outside_func = 0
    for stmt in ast.walk(ast.parse(replace_ipython_magics(nb_cell_code))):
        if isinstance(stmt, ast.FunctionDef) and not stmt.name.startswith("_"):
            for body_item in stmt.body:
                stmts_in_func += 1
        elif isinstance(stmt, ast.Module):
            for body_item in stmt.body:
                if not isinstance(body_item, ast.FunctionDef):
                    stmts_outside_func += 1
    return (
        0
        if stmts_outside_func + stmts_in_func == 0
        else (stmts_in_func / (stmts_outside_func + stmts_in_func)) * 100
    )

# %% ../nbs/indicators.ipynb 61
def in_func_pct(nb):
    nb_cell_code = "\n".join(
        [
            replace_ipython_magics(c["source"])
            for c in nb.cells
            if c["cell_type"] == "code"
        ]
    )
    if nb_cell_code == "":
        return np.nan
    return calc_ifp(nb_cell_code)

# %% ../nbs/indicators.ipynb 64
def markdown_code_pct(nb):
    md_cells = [c for c in nb.cells if c["cell_type"] == "markdown"]
    code_cells = [c for c in nb.cells if c["cell_type"] == "code"]
    num_code_cells = len(code_cells)
    if num_code_cells == 0:
        return np.nan
    num_md_cells = len(md_cells)
    return (
        100
        if num_code_cells == 0
        else (num_md_cells / (num_md_cells + num_code_cells)) * 100
    )

# %% ../nbs/indicators.ipynb 67
def total_code_len(nb):
    return sum([len(c["source"]) for c in nb.cells if c["cell_type"] == "code"])

# %% ../nbs/indicators.ipynb 70
def loc_per_md_section(nb):
    num_md_sections = len(
        [
            c["source"]
            for c in nb.cells
            if c["cell_type"] == "markdown" and c["source"].strip().startswith("#")
        ]
    )
    tcl = total_code_len(nb)
    if tcl == 0 or num_md_sections == 0:
        result = np.nan
    else:
        result = total_code_len(nb) / num_md_sections
    return result

# %% ../nbs/indicators.ipynb 73
indicator_funcs = {
    "calls_per_func_mean": calls_per_func_mean,
    "calls_per_func_median": calls_per_func_median,
    "tests_per_func_mean": tests_per_func_mean,
    "tests_func_coverage_pct": tests_func_coverage_pct,
    "in_func_pct": in_func_pct,
    "markdown_code_pct": markdown_code_pct,
    "loc_per_md_section": loc_per_md_section,
    "total_code_len": total_code_len,
}
